@@ -118,6 +118,9 @@ typedef struct NewGCMasterInfo {
   uintptr_t alive;
   uintptr_t ready;
   void **signal_fds;
+#ifdef POINTER_OWNERSHIP_CHECK
+  struct NewGC **places_gcs;
+#endif
   mzrt_rwlock *cangc;
   mzrt_sema *wait_sema;
 } NewGCMasterInfo;
@@ -136,6 +139,10 @@ typedef struct NewGC {
   Mark2_Proc  *mark_table;   /* the table of mark procs */
   Fixup2_Proc *fixup_table;  /* the table of repair procs */
   PageMap page_maps;
+#ifdef LOCK_PAGE_TABLES_FOR_DEBUG
+  mzrt_mutex *pagetable_lock;
+#endif
+
   /* All non-gen0 pages are held in the following structure. */
   struct mpage *gen1_pages[PAGE_TYPES];
 
@@ -238,6 +245,15 @@ typedef struct NewGC {
 
   Allocator *saved_allocator;
 
+#ifdef MZ_USE_PLACES
+  struct NewGC *parent_gc; /* parent for the purpose of reporting memory use */
+  intptr_t previously_reported_total; /* how much we previously reported to the parent */
+  mzrt_mutex *child_total_lock; /* lock on `child_gc_total' */
+#endif
+  intptr_t child_gc_total;
+
+  uintptr_t place_memory_limit; /* set to propagate a custodian limit from a parent place */  
+
 #if defined(GC_DEBUG_PAGES)
   FILE *GCVERBOSEFH;
 #endif
