@@ -466,6 +466,8 @@ Scheme_Object *(*scheme_alloc_char_string)(intptr_t size, mzchar fill);
 Scheme_Object *(*scheme_append_char_string)(Scheme_Object *, Scheme_Object *);
 mzchar *(*scheme_string_recase)(mzchar *s, int d, int len, int mode, int inplace, int *_len);
 Scheme_Object *(*scheme_make_vector)(intptr_t size, Scheme_Object *fill);
+Scheme_Double_Vector *(*scheme_alloc_flvector)(intptr_t size);
+Scheme_Vector *(*scheme_alloc_fxvector)(intptr_t size);
 Scheme_Object *(*scheme_make_integer_value)(intptr_t i);
 Scheme_Object *(*scheme_make_integer_value_from_unsigned)(uintptr_t i);
 Scheme_Object *(*scheme_make_integer_value_from_long_long)(mzlonglong i);
@@ -662,6 +664,7 @@ intptr_t (*scheme_output_tell)(Scheme_Object *port);
 intptr_t (*scheme_tell_line)(Scheme_Object *port);
 intptr_t (*scheme_tell_column)(Scheme_Object *port);
 void (*scheme_tell_all)(Scheme_Object *port, intptr_t *line, intptr_t *col, intptr_t *pos);
+void (*scheme_set_port_location)(int argc, Scheme_Object **argv);
 void (*scheme_count_lines)(Scheme_Object *port);
 void (*scheme_close_input_port)(Scheme_Object *port);
 void (*scheme_close_output_port)(Scheme_Object *port);
@@ -937,6 +940,7 @@ void (*scheme_signal_received_at)(void *);
 void *(*scheme_get_signal_handle)();
 intptr_t (*scheme_char_strlen)(const mzchar *s);
 Scheme_Object *(*scheme_stx_extract_marks)(Scheme_Object *stx);
+int (*scheme_get_place_id)(void);
 Scheme_Hash_Table *(*scheme_get_place_table)(void);
 void *(*scheme_register_process_global)(const char *key, void *val);
 Scheme_Object *(*scheme_malloc_key)(void);
