@@ -336,6 +336,8 @@
 #define scheme_append_char_string (scheme_extension_table->scheme_append_char_string)
 #define scheme_string_recase (scheme_extension_table->scheme_string_recase)
 #define scheme_make_vector (scheme_extension_table->scheme_make_vector)
+#define scheme_alloc_flvector (scheme_extension_table->scheme_alloc_flvector)
+#define scheme_alloc_fxvector (scheme_extension_table->scheme_alloc_fxvector)
 #define scheme_make_integer_value (scheme_extension_table->scheme_make_integer_value)
 #define scheme_make_integer_value_from_unsigned (scheme_extension_table->scheme_make_integer_value_from_unsigned)
 #define scheme_make_integer_value_from_long_long (scheme_extension_table->scheme_make_integer_value_from_long_long)
@@ -473,6 +475,7 @@
 #define scheme_tell_line (scheme_extension_table->scheme_tell_line)
 #define scheme_tell_column (scheme_extension_table->scheme_tell_column)
 #define scheme_tell_all (scheme_extension_table->scheme_tell_all)
+#define scheme_set_port_location (scheme_extension_table->scheme_set_port_location)
 #define scheme_count_lines (scheme_extension_table->scheme_count_lines)
 #define scheme_close_input_port (scheme_extension_table->scheme_close_input_port)
 #define scheme_close_output_port (scheme_extension_table->scheme_close_output_port)
@@ -670,6 +673,7 @@
 #define scheme_get_signal_handle (scheme_extension_table->scheme_get_signal_handle)
 #define scheme_char_strlen (scheme_extension_table->scheme_char_strlen)
 #define scheme_stx_extract_marks (scheme_extension_table->scheme_stx_extract_marks)
+#define scheme_get_place_id (scheme_extension_table->scheme_get_place_id)
 #define scheme_get_place_table (scheme_extension_table->scheme_get_place_table)
 #define scheme_register_process_global (scheme_extension_table->scheme_register_process_global)
 #define scheme_malloc_key (scheme_extension_table->scheme_malloc_key)
